# -*- coding: utf-8 -*-
# ===============LICENSE_START=======================================================
# Acumos Apache-2.0
# ===================================================================================
# Copyright (C) 2017-2018 AT&T Intellectual Property & Tech Mahindra. All rights reserved.
# ===================================================================================
# This Acumos software file is distributed by AT&T and Tech Mahindra
# under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# This file is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============LICENSE_END=========================================================
"""
Provides metadata generation utilities
"""
import importlib
import sys
from types import ModuleType
from pkg_resources import get_distribution, DistributionNotFound
from os.path import basename, normpath, sep as pathsep

from acumos.exc import AcumosError


BUILTIN_MODULE_NAMES = set(sys.builtin_module_names)
PACKAGE_DIRS = {'site-packages', 'dist-packages'}
SCHEMA_VERSION = '0.4.0'
_SCHEMA = "acumos.schema.model:{}".format(SCHEMA_VERSION)

# known package mappings because Python packaging is madness
_REQ_MAP = {
    'sklearn': 'scikit-learn',
    'PIL': 'Pillow'
}


class Requirements(object):
    '''
    A collection of optional user-provided Python requirement metadata

    Parameters
    ----------
    reqs : Sequence[str]
        A sequence of pip-installable Python package names
    req_map : Dict[str, str] or Dict[module, str]
        A corrective mapping of Python modules to pip-installable package names. For example
        `req_map={'sklearn': 'scikit-learn'}` or `import sklearn; req_map={sklearn: 'scikit-learn'}`.
    packages : Sequence[str]
        A sequence of paths to directories containing Python packages (i.e. Python packages). The
        directories will be copied along with your model and added to the PYTHONPATH at runtime.
    '''

    __slots__ = ('reqs', 'req_map', 'packages')

    def __init__(self, reqs=None, req_map=None, packages=None):
        self.reqs = set() if reqs is None else _safe_set(reqs)
        self.req_map = dict() if req_map is None else req_map.copy()
        self.req_map.update(_REQ_MAP)
        self.packages = set() if packages is None else _safe_set(packages)

    @property
    def package_names(self):
        return set(map(basename, self.packages))


def _safe_set(input_):
    '''Safely returns a set from input'''
    return {input_, } if isinstance(input_, str) else set(input_)


def create_model_meta(model, name, requirements, encoding='protobuf'):
    '''Returns a model metadata dictionary'''
    return {'schema': _SCHEMA,
            'runtime': _create_runtime(requirements, encoding),
            'name': name,
            'methods': {name: {'input': f.input_type.__name__,
                               'output': f.output_type.__name__,
                               'description': f.description} for name, f in model.methods.items()}}


def _create_runtime(requirements, encoding='protobuf'):
    '''Returns a runtime dict'''
    req_set = _create_requirement_set(requirements)
    req_tuples = _req_iter(req_set, requirements.package_names)
    return {'name': 'python',
            'encoding': encoding,
            'version': '.'.join(map(str, sys.version_info[:3])),
            'dependencies': _create_requirements(req_tuples)}


def _create_requirement_set(requirements):
    '''Returns a set of requirement names'''
    req_names = (n for n in map(_get_requirement_name, requirements.reqs) if not _in_stdlib(n))
    req_map = {_get_requirement_name(k): v for k, v in requirements.req_map.items()}
    mapped_reqs = {req_map.get(r, r) for r in req_names}
    return mapped_reqs


def _get_requirement_name(req):
    '''Returns the str name of the requirement'''
    if isinstance(req, ModuleType):
        name = req.__name__
    elif isinstance(req, str):
        name = req
    else:
        raise AcumosError("Requirement {} is invalid; must be ModuleType or string".format(req))
    return name


def _req_iter(req_names, package_names):
    '''Yields (package, version) tuples'''
    for req_name in req_names:
        try:
            yield str(get_distribution(req_name).as_requirement()).split('==')
        except DistributionNotFound:
            if req_name not in package_names:
                raise AcumosError("Module {} was detected as a dependency, but not found as a pip installed package. Use acumos.session.Requirements to declare custom packages or map module names to pip-installable names (e.g. Requirements(req_map=dict(cv2='opencv-python')) )".format(req_name))


def _in_stdlib(module_name):
    '''Returns True if the package name is part of the standard library (not airtight)'''
    if module_name in BUILTIN_MODULE_NAMES:
        return True

    try:
        install_path = importlib.import_module(module_name).__file__
    except ImportError:
        return False

    if not install_path.startswith(sys.prefix):
        return False

    dirs = set(normpath(install_path).split(pathsep))
    return not PACKAGE_DIRS & dirs


def _create_requirements(req_tuples):
    '''Returns a dict containing model implementation metadata'''
    return {
        'pip': {
            'indexes': [],
            'requirements': [{'name': n, 'version': v} for n, v in req_tuples]
        },
        'conda': {
            'channels': [],
            'requirements': []
        }
    }



# -*- coding: utf-8 -*-
# ===============LICENSE_START=======================================================
# Acumos Apache-2.0
# ===================================================================================
# Copyright (C) 2017-2018 AT&T Intellectual Property & Tech Mahindra. All rights reserved.
# ===================================================================================
# This Acumos software file is distributed by AT&T and Tech Mahindra
# under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# This file is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============LICENSE_END=========================================================
"""
Provides a Acumos session for pushing and dumping models
"""
import random
import string
import shutil
import json
import contextlib
import requests
import fnmatch
from tempfile import TemporaryDirectory
from os import walk, mkdir
from os.path import dirname, isdir, expanduser, relpath, basename, join as path_join
from pathlib import Path

from acumos.pickler import AcumosContextManager, dump_model
from acumos.metadata import create_model_meta, Requirements
from acumos.utils import dump_artifact, get_qualname
from acumos.exc import AcumosError
from acumos.protogen import model2proto, compile_protostr
from acumos.logging import get_logger
from acumos.modeling import Model
from acumos.auth import get_jwt, clear_jwt


logger = get_logger(__name__)


class AcumosSession(object):
    '''
    A session that enables on-boarding models to Acumos

    Parameters
    ----------
    push_api : str
        The full URL to the Acumos on-boarding server upload API
    auth_api : str
        The full URL to the Acumos on-boarding server authentication API
    '''

    def __init__(self, push_api=None, auth_api=None):
        self.push_api = push_api
        self.auth_api = auth_api

    def push(self, model, name, requirements=None, extra_headers=None):
        '''
        Pushes a model to Acumos

        Parameters
        ----------
        model : ``acumos.modeling.Model``
            An Acumos model instance
        name : str
            The name of your model
        requirements : ``acumos.session.Requirements``, optional
            Additional Python dependencies that you can optionally specify
        extra_headers : dict, optional
            Additonal HTTP headers included in the POST to the Acumos onboarding server
        '''
        _assert_valid_input(model, requirements)
        _assert_valid_apis(push_api=self.push_api, auth_api=self.auth_api)

        with _dump_model(model, name, requirements) as dump_dir:
            _push_model(dump_dir, self.push_api, self.auth_api, extra_headers=extra_headers)

    def dump(self, model, name, outdir, requirements=None):
        '''
        Creates a directory located at ``outdir/name`` containing Acumos model artifacts

        Parameters
        ----------
        model : ``acumos.modeling.Model``
            An Acumos model instance
        name : str
            The name of your model
        outdir : str
            The directory or folder to save your model .zip to
        requirements : ``acumos.session.Requirements``, optional
            Additional Python dependencies that you can optionally specify
        '''
        _assert_valid_input(model, requirements)
        with _dump_model(model, name, requirements) as dump_dir:
            _copy_dir(dump_dir, expanduser(outdir), name)


def _assert_valid_input(model, requirements):
    '''Raises AcumosError if inputs are invalid'''
    if not isinstance(model, Model):
        raise AcumosError("Input `model` must be of type {}".format(get_qualname(Model)))

    if requirements is not None and not isinstance(requirements, Requirements):
        raise AcumosError("Input `requirements` must be of type {}".format(get_qualname(Requirements)))


def _assert_valid_apis(**apis):
    '''Raises AcumosError if api are invalid'''
    for param, api in apis.items():
        if api is None:
            raise AcumosError("AcumosSession.push requires that the API for `{}` be provided".format(param))

        if not api.startswith('https'):
            logger.warning("Provided `{}` API {} does not begin with 'https'. Your password and token are visible in plaintext!".format(param, api))


def _push_model(dump_dir, push_api, auth_api, max_tries=2, extra_headers=None):
    '''Pushes a model to the Acumos server'''
    with open(path_join(dump_dir, 'model.zip'), 'rb') as model, \
            open(path_join(dump_dir, 'metadata.json')) as meta, \
            open(path_join(dump_dir, 'model.proto')) as proto:

        files = {'model': ('model.zip', model, 'application/zip'),
                 'metadata': ('metadata.json', meta, 'application/json'),
                 'schema': ('model.proto', proto, 'application/text')}

        _post_model(files, push_api, auth_api, 1, max_tries, extra_headers)


def _post_model(files, push_api, auth_api, tries, max_tries, extra_headers):
    '''Attempts to post the model to Acumos'''
    headers = {'Authorization': get_jwt(auth_api)}
    if extra_headers is not None:
        headers.update(extra_headers)

    r = requests.post(push_api, files=files, headers=headers)

    if r.status_code == 201:
        logger.info("Model successfully pushed to {}".format(push_api))
    elif r.status_code == 401:
        clear_jwt()
        if tries == max_tries:
            raise AcumosError("Authentication succeeded but authorization failed: {}".format(r.text))
        else:
            logger.warning('Authorization failed. Trying again')
            _post_model(files, push_api, auth_api, tries + 1, max_tries, extra_headers)
    else:
        raise AcumosError("Model upload failed: {}".format(r.text))


@contextlib.contextmanager
def _dump_model(model, name, requirements=None):
    '''Generates model artifacts and serializes the model'''
    requirements = Requirements() if requirements is None else requirements

    with TemporaryDirectory() as rootdir:

        model_dir = path_join(rootdir, 'model')
        mkdir(model_dir)

        with AcumosContextManager(model_dir) as c:

            with open(c.build_path('model.pkl'), 'wb') as f:
                dump_model(model, f)

            # generate protobuf definition
            proto_pkg = c.parameters['protobuf_package'] = _random_string()
            protostr = model2proto(model, proto_pkg)
            dump_artifact(rootdir, 'model.proto', data=protostr, module=None, mode='w')

            # generate protobuf source code
            module_name = 'model'
            proto_dir = c.create_subdir(path_join('scripts', 'acumos_gen', proto_pkg))
            compile_protostr(protostr, proto_pkg, module_name, proto_dir)

            # generate model metadata
            requirements.reqs.update(c.modules)
            metadata = create_model_meta(model, name, requirements)
            dump_artifact(rootdir, 'metadata.json', data=metadata, module=json, mode='w')

            # bundle user-provided scripts
            user_provided = c.create_subdir(path_join('scripts', 'user_provided'))
            Path(user_provided, '.keep').touch()  # may resolve pruning issues when unzipping

            scripts = _gather_scripts(requirements.packages)
            _copy_scripts(c, scripts)

        shutil.make_archive(model_dir, 'zip', model_dir)  # create zip at same level as parent
        shutil.rmtree(model_dir)  # clean up model directory

        yield rootdir


def _copy_dir(src_dir, outdir, name):
    '''Copies a directory to a new location'''
    dst_path = path_join(outdir, name)
    if isdir(dst_path):
        raise AcumosError("Model {} has already been dumped".format(dst_path))
    shutil.copytree(src_dir, dst_path)


def _copy_scripts(context, scripts, prefix=path_join('scripts', 'user_provided')):
    '''Moves all gathered scripts to the context directory'''
    for script_relpath, script_abspath in scripts:
        context_subdir = path_join(prefix, dirname(script_relpath))
        context_absdir = context.create_subdir(context_subdir, exist_ok=True)
        shutil.copy(script_abspath, context_absdir)


def _gather_scripts(packages):
    '''Yields (relpath, abspath) tuples of Python from a sequence of directories'''
    for path in packages:
        path = expanduser(path)
        if not isdir(path):
            raise AcumosError("Path {} is not a directory".format(path))

        for root, dirnames, filenames in walk(path):
            for filename in fnmatch.filter(filenames, '*.py'):
                script_abspath = path_join(root, filename)
                script_relpath = path_join(basename(path), relpath(script_abspath, path))
                yield script_relpath, script_abspath


def _random_string(length=32):
    '''Returns a random string containing ascii characters'''
    return ''.join(random.choice(string.ascii_letters) for _ in range(length))



# -*- coding: utf-8 -*-
# ===============LICENSE_START=======================================================
# Acumos Apache-2.0
# ===================================================================================
# Copyright (C) 2017-2018 AT&T Intellectual Property & Tech Mahindra. All rights reserved.
# ===================================================================================
# This Acumos software file is distributed by AT&T and Tech Mahindra
# under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# This file is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============LICENSE_END=========================================================
"""
Provides custom pickle utilities
"""
import json
import inspect
import contextlib
import tempfile
from os import makedirs
from os.path import basename, isdir, isfile, join as path_join
from copy import deepcopy
from functools import partial
from typing import GenericMeta, Dict, List

import dill

from acumos.modeling import _is_namedtuple, create_namedtuple, Empty
from acumos.exc import AcumosError


_BLACKLIST = {'builtins', '__main__'}
_DEFAULT = 'default'

_contexts = dict()

dump_model = partial(dill.dump, recurse=True)
dumps_model = partial(dill.dumps, recurse=True)
load_model = dill.load
loads_model = dill.loads


def _save_annotation(pickler, obj):
    '''Workaround for dill annotation serialization bug'''
    if obj.__origin__ in (Dict, List):
        # recursively save object
        t = obj.__origin__
        args = obj.__args__
        pickler.save_reduce(_load_annotation, (t, args), obj=obj)
    else:
        # eventually hit base type, then use stock pickling logic. temp revert prevents infinite recursion
        t = obj
        args = None
        with _revert_dispatch(GenericMeta):
            pickler.save_reduce(_load_annotation, (t, args), obj=obj)


def _load_annotation(t, args):
    '''Workaround for dill annotation serialization bug'''
    if t is Dict and args is not None:
        return Dict[args[0], args[1]]
    elif t is List and args is not None:
        return List[args[0]]
    else:
        return t


def _save_namedtuple(pickler, obj):
    '''Workaround for dill NamedTuple serialization bug'''
    pickler.save_reduce(_load_namedtuple, (obj.__name__, obj._field_types), obj=obj)


def _load_namedtuple(name, field_types):
    '''Workaround for dill NamedTuple serialization bug'''
    return create_namedtuple(name, [(k, v) for k, v in field_types.items()])


@contextlib.contextmanager
def _revert_dispatch(t):
    '''Temporarily removes a type from the dispatch table to use existing serialization logic'''
    f = dill.Pickler.dispatch.pop(t)
    try:
        yield
    finally:
        dill.Pickler.dispatch[t] = f


def _save_keras(pickler, obj):
    '''Serializes a keras model to a context directory'''
    from keras.backend import backend

    context = get_context()
    model_subdir_abspath = context.create_subdir()  # /path/to/context/root/abc123

    model_name = 'model.h5'
    model_subpath = path_join(basename(model_subdir_abspath), model_name)  # abc123/model.h5
    model_abspath = path_join(context.abspath, model_subpath)  # /path/to/context/root/abc123/model.h5

    obj.save(model_abspath)
    context.add_module('h5py')  # needed for keras model serialization
    context.add_module(backend())  # adds name of active keras backend

    # check for custom or contrib layer modules
    layer_modules = (_get_top_module(l)[1] for l in obj.layers)
    special_layers = tuple((l.__class__, m) for l, m in zip(obj.layers, layer_modules) if m != 'keras')
    for _, module in special_layers:
        context.add_module(module)

    # store subpath because context root can change, and special layer classes to have them imported before load
    pickler.save_reduce(_load_keras, (model_subpath, special_layers), obj=obj)


def _load_keras(model_subpath, special_classes):
    '''Loads a keras model from a context subdirectory'''
    from keras.models import load_model as load_keras_model

    context = get_context()
    model_path = context.build_path(model_subpath)  # /different/path/to/context/root/abc123/model.h5
    model = load_keras_model(model_path)
    return model


def _save_tf_tensor(pickler, tensor):
    '''Saves a TensorFlow tensor object'''
    import tensorflow as tf
    pickler.save_reduce(_load_tf_tensor, (tensor.name, tensor.graph, tf.get_default_session()), obj=tensor)


def _load_tf_tensor(name, graph, session):
    '''Loads a TensorFlow tensor object'''
    return graph.get_tensor_by_name(name)


def _save_tf_operation(pickler, op):
    '''Saves a TensorFlow operation object'''
    import tensorflow as tf
    pickler.save_reduce(_load_tf_operation, (op.name, op.graph, tf.get_default_session()), obj=op)


def _load_tf_operation(name, graph, session):
    '''Loads a TensorFlow operation object'''
    return graph.get_operation_by_name(name)


def _save_tf_session(pickler, session):
    '''Saves a TensorFlow session'''
    import tensorflow as tf

    context = get_context()
    model_subdir_abspath = context.create_subdir()  # /path/to/context/root/abc123

    model_name = 'model.ckpt'
    model_subpath = path_join(basename(model_subdir_abspath), model_name)  # abc123/model.ckpt
    model_abspath = path_join(context.abspath, model_subpath)  # /path/to/context/root/abc123/model.ckpt

    interactive = isinstance(session, tf.InteractiveSession)

    # graph is included in save_reduce in order to make sure that it is restored before trying
    # to restore the session, which would result in an error due to missing tensors / operations
    graph = session.graph

    saver = tf.train.Saver()
    saver.save(session, model_abspath)
    pickler.save_reduce(_load_tf_session, (model_subpath, interactive, graph), obj=session)


def _load_tf_session(model_subpath, interactive, graph):
    '''Loads a TensorFlow session'''
    import tensorflow as tf

    context = get_context()
    model_path = context.build_path(model_subpath)  # /different/path/to/context/root/abc123/model.ckpt

    with graph.as_default():
        sess = tf.InteractiveSession() if interactive else tf.Session()
        saver = tf.train.Saver()
        saver.restore(sess, model_path)
    return sess


def _save_tf_graph(pickler, graph):
    '''Saves a TensorFlow graph'''
    import tensorflow as tf

    context = get_context()
    model_subdir_abspath = context.create_subdir()  # /path/to/context/root/abc123

    graph_name = 'graph.meta'
    graph_subpath = path_join(basename(model_subdir_abspath), graph_name)
    graph_abspath = path_join(context.abspath, graph_subpath)

    with graph.as_default():
        tf.train.export_meta_graph(graph_abspath)  # exports default graph
    pickler.save_reduce(_load_tf_graph, (graph_subpath, ), obj=graph)


def _load_tf_graph(graph_subpath):
    '''Loads a TensorFlow graph'''
    import tensorflow as tf

    context = get_context()
    graph_path = context.build_path(graph_subpath)

    graph = tf.Graph()
    with graph.as_default():
        tf.train.import_meta_graph(graph_path)
    return graph


dill.Pickler.dispatch[GenericMeta] = _save_annotation


_CUSTOM_DISPATCH = {
    'keras.engine.training.Model': _save_keras,
    'tensorflow.python.framework.ops.Tensor': _save_tf_tensor,
    'tensorflow.python.framework.ops.Operation': _save_tf_operation,
    'tensorflow.python.client.session.BaseSession': _save_tf_session,
    'tensorflow.python.framework.ops.Graph': _save_tf_graph,
}


@contextlib.contextmanager
def _patch_dill():
    '''Temporarily patches the dill Pickler dispatch table to support custom serialization within a context'''
    try:
        dispatch = dill.Pickler.dispatch
        dill.Pickler.dispatch = deepcopy(dispatch)

        pickler_save = dill.Pickler.save

        def wrapped_save(pickler, obj, save_persistent_id=True):
            '''Hook that intercepts objects about to be saved'''
            _catch_object(obj)

            if _is_namedtuple(obj) and obj is not Empty:
                _save_namedtuple(pickler, obj)
            else:
                pickler_save(pickler, obj, save_persistent_id)

        dill.Pickler.save = wrapped_save
        yield
    finally:
        dill.Pickler.dispatch = dispatch
        dill.Pickler.save = pickler_save


def _catch_object(obj):
    '''Inspects object and executes custom serialization / bookkeeping logic'''

    # dynamically extend dispatch table to prevent unnecessary imports / dependencies
    obj_type = obj if inspect.isclass(obj) else type(obj)
    if obj_type not in dill.Pickler.dispatch:
        for path in _get_mro_paths(obj_type):
            if path in _CUSTOM_DISPATCH:
                dill.Pickler.dispatch[obj_type] = _CUSTOM_DISPATCH[path]

    mod, name = _get_top_module(obj)
    if name is not None and name not in _BLACKLIST:
        # this check is the current solution for discarding unwanted scripts like __main__
        if mod.__package__:
            context = get_context()
            context.add_module(name)


def _get_top_module(obj):
    '''Returns the top-level module for a given object'''
    mod = inspect.getmodule(obj)
    if mod is not None:
        base, _, _ = mod.__name__.partition('.')
    else:
        base = None
    return mod, base


def _get_mro_paths(type_):
    '''Yields import path string for each entry in `getmro`'''
    for t in inspect.getmro(type_):
        yield "{}.{}".format(t.__module__, t.__name__)


class AcumosContext(object):
    '''Represents a workspace for a model that is being dumped'''

    def __init__(self, root_dir):
        if not isdir(root_dir):
            raise AcumosError("AcumosContext root directory {} does not exist".format(root_dir))

        self._root_dir = root_dir
        self._modules = {'acumos', 'dill'}
        self._params_path = path_join(root_dir, 'context.json')
        self.parameters = self._load_params()

    def create_subdir(self, name=None, exist_ok=False):
        '''Creates a new directory within the context root and returns the absolute path'''
        if name is None:
            tdir = tempfile.mkdtemp(dir=self._root_dir)
        else:
            tdir = path_join(self._root_dir, name)
            makedirs(tdir, exist_ok=exist_ok)
        return tdir

    def build_path(self, *args):
        '''Returns an absolute path starting from the context root'''
        return path_join(self._root_dir, *args)

    def add_module(self, module):
        '''Adds a module to the context module set'''
        self._modules.add(module)

    @property
    def abspath(self):
        return self._root_dir

    @property
    def basename(self):
        return basename(self._root_dir)

    @property
    def modules(self):
        return set(self._modules)

    def _load_params(self):
        '''Returns a parameters dict'''
        if isfile(self._params_path):
            with open(self._params_path) as f:
                return json.load(f)
        else:
            return dict()

    def save_params(self):
        '''Saves a parameters json file within the context root'''
        with open(self._params_path, 'w') as f:
            json.dump(self.parameters, f)


@contextlib.contextmanager
def AcumosContextManager(rootdir=None, name=_DEFAULT):
    '''Context manager that provides a AcumosContext object'''
    with _patch_dill():
        if name in _contexts:
            raise AcumosError("AcumosContext '{}' has already been created. Use `get_context` to access it.".format(name))
        try:
            with _DirManager(rootdir) as rootdir:
                context = AcumosContext(rootdir)
                _contexts[name] = context
                yield context
                context.save_params()
        finally:
            del _contexts[name]


@contextlib.contextmanager
def _DirManager(dir_=None):
    '''Wrapper that passes dir_ through or creates a temporary directory'''
    if dir_ is not None:
        if not isdir(dir_):
            raise AcumosError("Provided AcumosContext rootdir {} does not exist".format(dir_))
        yield dir_
    else:
        with tempfile.TemporaryDirectory() as tdir:
            yield tdir


def get_context(name=_DEFAULT):
    '''Returns an existing AcumosContext'''
    if name in _contexts:
        return _contexts[name]
    else:
        raise AcumosError("AcumosContext '{}' has not been created".format(name))



# -*- coding: utf-8 -*-
# ===============LICENSE_START=======================================================
# Acumos Apache-2.0
# ===================================================================================
# Copyright (C) 2017-2018 AT&T Intellectual Property & Tech Mahindra. All rights reserved.
# ===================================================================================
# This Acumos software file is distributed by AT&T and Tech Mahindra
# under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# This file is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============LICENSE_END=========================================================
"""
Provides model wrapping utilities
"""
import sys
import os
import re
import inspect
import contextlib

from acumos.exc import AcumosError


def _load_module_py33(fullname, path):
    '''Imports and returns a module from path for Python 3.3-3.4'''
    from importlib.machinery import SourceFileLoader
    module = SourceFileLoader(fullname, path).load_module()
    return module


def _load_module_py35(fullname, path):
    '''Imports and returns a module from path for Python 3.5+'''
    import importlib.util
    spec = importlib.util.spec_from_file_location(fullname, path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def load_module(fullname, path):
    '''Imports and returns a module from path'''
    ver_info = sys.version_info
    if (3, 3) <= ver_info < (3, 5):
        return _load_module_py33(fullname, path)
    elif (3, 5) <= ver_info:
        return _load_module_py35(fullname, path)
    else:
        raise AcumosError("Attempted to import Python module from path, but Python {} is not supported".format(ver_info))


def load_artifact(*path, module, mode):
    '''Artifact loader helper'''
    with open(os.path.join(*path), mode) as f:
        return module.load(f)


def dump_artifact(*path, data, module, mode):
    '''Artifact saver helper'''
    with open(os.path.join(*path), mode) as f:
        if module is None:
            f.write(data)
        else:
            module.dump(data, f)


def sanitize_field(name):
    return re.sub(r"\W+", '_', name)


def get_qualname(o):
    if inspect.isclass(o):
        return "{}.{}".format(o.__module__, o.__name__)
    else:
        return get_qualname(o.__class__)


@contextlib.contextmanager
def reraise(prefix, prefix_args):
    '''Reraises an exception with a more informative prefix'''
    try:
        yield
    except AcumosError as e:
        raise AcumosError("{}: {}".format(prefix.format(*prefix_args), e)).with_traceback(e.__traceback__)



# -*- coding: utf-8 -*-
# ===============LICENSE_START=======================================================
# Acumos Apache-2.0
# ===================================================================================
# Copyright (C) 2017-2018 AT&T Intellectual Property & Tech Mahindra. All rights reserved.
# ===================================================================================
# This Acumos software file is distributed by AT&T and Tech Mahindra
# under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# This file is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============LICENSE_END=========================================================
"""
Provides logging utilities
"""
import logging

import acumos


_handler = logging.StreamHandler()
_handler.setFormatter(logging.Formatter('[%(levelname)s] %(name)s : %(message)s'))

_root = logging.getLogger(acumos.__name__)
_root.setLevel(logging.INFO)
_root.handlers = [_handler, ]
_root.propagate = False


def get_logger(name):
    '''Returns a logger object'''
    return logging.getLogger(name)
